// file      : structure-msxml.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006 Code Synthesis Tools CC
// license   : GNU GPL v2; see accompanying LICENSE file

#include <iostream>
#include <fstream>

#include <msxml2.h>
#include <comdef.h>

#include "time.hxx"

using namespace std;

// Declare the 'smart pointer' type to simplify COM handling.
//
_COM_SMARTPTR_TYPEDEF(ISAXXMLReader, __uuidof(ISAXXMLReader));
_COM_SMARTPTR_TYPEDEF(IXMLDOMSchemaCollection,
                      __uuidof(IXMLDOMSchemaCollection));

class ContentHandler: public ISAXContentHandler
{
public:
  // Assuming that our handler is not a COM object, always
  // returning 0 in following three functions.
  //
  long __stdcall QueryInterface(const struct _GUID &,void ** ) {return 0;}
  unsigned long __stdcall AddRef(void) { return 1;}
  unsigned long __stdcall Release(void) { return 1;}

  virtual HRESULT __stdcall putDocumentLocator(
    /* [in] */ ISAXLocator *pLocator) { return S_OK; }

  virtual HRESULT __stdcall startDocument( void) { return S_OK; }

  virtual HRESULT __stdcall endDocument( void) { return S_OK; }

  virtual HRESULT __stdcall startPrefixMapping(
    /* [in] */ const wchar_t* pwchPrefix,
    /* [in] */ int cchPrefix,
    /* [in] */ const wchar_t* pwchUri,
    /* [in] */ int cchUri) { return S_OK; }

  virtual HRESULT __stdcall endPrefixMapping(
    /* [in] */ const wchar_t *pwchPrefix,
    /* [in] */ int cchPrefix) { return S_OK; }

  virtual HRESULT __stdcall startElement(
    /* [in] */ const wchar_t *pwchNamespaceUri,
    /* [in] */ int cchNamespaceUri,
    /* [in] */ const wchar_t *pwchLocalName,
    /* [in] */ int cchLocalName,
    /* [in] */ const wchar_t *pwchRawName,
    /* [in] */ int cchRawName,
    /* [in] */ ISAXAttributes __RPC_FAR *pAttributes) { return S_OK; }

  virtual HRESULT __stdcall endElement(
    /* [in] */ const wchar_t *pwchNamespaceUri,
    /* [in] */ int cchNamespaceUri,
    /* [in] */ const wchar_t *pwchLocalName,
    /* [in] */ int cchLocalName,
    /* [in] */ const wchar_t *pwchRawName,
    /* [in] */ int cchRawName) { return S_OK; }

  virtual HRESULT __stdcall characters(
    /* [in] */ const wchar_t *pwchChars,
    /* [in] */ int cchChars) { return S_OK; }

  virtual HRESULT __stdcall ignorableWhitespace(
    /* [in] */ const wchar_t *pwchChars,
    /* [in] */ int cchChars) { return S_OK; }

  virtual HRESULT __stdcall processingInstruction(
    /* [in] */ const wchar_t *pwchTarget,
    /* [in] */ int cchTarget,
    /* [in] */ const wchar_t *pwchData,
    /* [in] */ int cchData) { return S_OK; }

  virtual HRESULT __stdcall skippedEntity(
    /* [in] */ const wchar_t *pwchName,
    /* [in] */ int cchName) { return S_OK; }
};

class ErrorHandler: public ISAXErrorHandler
{
public:
  long __stdcall QueryInterface(const struct _GUID &,void ** ) {return 0;}
  unsigned long __stdcall AddRef(void) { return 1;}
  unsigned long __stdcall Release(void) { return 1;}

  virtual HRESULT STDMETHODCALLTYPE error(
    /* [in] */ ISAXLocator *pLocator,
    /* [in] */ const wchar_t *pwchErrorMessage,
    /* [in] */ HRESULT hrErrorCode)
  {
    int line, column;
    pLocator->getColumnNumber (&column);
    pLocator->getLineNumber (&line);

    char str[1024];
    wcstombs (str, pwchErrorMessage, 1023);
    str[1023] = '\0';

    cerr << line << ":" << column << ": error: "
         << str << endl;
    return S_OK;
  }

  virtual HRESULT STDMETHODCALLTYPE fatalError(
    /* [in] */ ISAXLocator *pLocator,
    /* [in] */ const wchar_t *pwchErrorMessage,
    /* [in] */ HRESULT hrErrorCode)
  {
    return error (pLocator, pwchErrorMessage, hrErrorCode);
  }

  virtual HRESULT STDMETHODCALLTYPE ignorableWarning(
    /* [in] */ ISAXLocator *pLocator,
    /* [in] */ const wchar_t *pwchErrorMessage,
    /* [in] */ HRESULT hrErrorCode)
  {
    return S_OK;
  }
};


class Stream: public ISequentialStream
{
public:
  Stream (const char* buf, size_t size)
      : buf_ (buf), size_ (size), pos_ (0)
  {
  }

  virtual HRESULT __stdcall
  Read(void* pv, ULONG cb, ULONG* pcbRead)
  {
    size_t s (min (size_ - pos_, cb));
    memcpy (pv, buf_ + pos_, s);
    pos_ += s;
    *pcbRead = static_cast<ULONG> (s);
    return S_OK;
  }

  virtual HRESULT __stdcall
  Write(const void __RPC_FAR *pv, ULONG cb, ULONG __RPC_FAR *pcbWritten)
  {
    return S_FALSE;
  }

  virtual long __stdcall
  QueryInterface(const struct _GUID &riid,void **ppvObject)
  {
    // We have to implement this, because we pass this as an IUnknown but
    // it needs an IStream interface.
    //
    if(riid == __uuidof(ISequentialStream))
    {
      *ppvObject = reinterpret_cast<void*>(this);
      return S_OK;
    }

    return E_NOINTERFACE;
  }

  virtual unsigned long __stdcall
  AddRef() { return 1; }

  virtual unsigned long __stdcall
  Release() { return 1; }

private:
  const char* buf_;
  size_t size_;
  size_t pos_;
};

int
main (int argc, char* argv[])
{
  if (argc != 2)
  {
    cerr << "usage: " << argv[0] << " strcuture.xml" << endl;
    return 1;
  }

  try
  {
    ifstream ifs;
    ifs.exceptions (ios_base::failbit);
    ifs.open (argv[1], ios::in | ios::ate);

    size_t size (ifs.tellg ());
    ifs.seekg (0, ios::beg);

    char* buf = new char[size];
    ifs.read (buf, size);

    cerr << "document size: " << size << " bytes" << endl;

    CoInitialize (0);

    ISAXXMLReaderPtr sax;

    sax.CreateInstance ("Msxml2.SAXXMLReader.6.0");

    if(sax.GetInterfacePtr () == 0)
      sax.CreateInstance ("Msxml2.SAXXMLReader.4.0");

    if(sax.GetInterfacePtr () == 0)
      sax.CreateInstance ("Msxml2.SAXXMLReader.3.0");

    if(sax.GetInterfacePtr () == 0)
      sax.CreateInstance (__uuidof(ISAXXMLReader));

    if(sax.GetInterfacePtr () == 0)
    {
      return 1;
    }

    ContentHandler handler;
    sax->putContentHandler (&handler);

    ErrorHandler error_handler;
    sax->putErrorHandler (&error_handler);

    sax->putFeature (L"schema-validation", true);


    IXMLDOMSchemaCollectionPtr cache;

    cache.CreateInstance ("Msxml2.XMLSchemaCache.6.0");

    if(cache.GetInterfacePtr () == 0)
      cache.CreateInstance ("Msxml2.XMLSchemaCache.4.0");

    if(cache.GetInterfacePtr () == 0)
      cache.CreateInstance ("Msxml2.XMLSchemaCache.3.0");

    if(cache.GetInterfacePtr () == 0)
    {
      cerr << "Unable to create MSXML Schema cache" << endl;
      return 1;
    }

    {
      _variant_t v;
      v.SetString ("structure.xsd");
      if (cache->add (
            L"http://www.codesynthesis.com/xmlns/xsdbench/structure", v)
          != S_OK)
      {
        cerr << "unable to add schema to the cache" << endl;
        return 1;
      }
    }

    {
      _variant_t v (cache.GetInterfacePtr ());
      sax->putProperty (L"schemas", v);
    }

    OS::Time start;

    for (unsigned long i (0); i < 10000; ++i)
    {
      Stream s (buf, size);
      _variant_t v (&s);
      if (sax->parse (v) != S_OK)
      {
        cerr << "parsing failed" << endl;
        return 1;
      }
    }

    OS::Time end;

    cerr << "time: " << end - start << endl;

    cache.Release ();
    sax.Release();
    CoUninitialize();
    delete[] buf;
  }
  catch (std::ios_base::failure const&)
  {
    wcerr << "io failure" << endl;
    return 1;
  }
}
