// file      : time.hxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2006 Code Synthesis Tools CC
// license   : GNU GPL v2; see accompanying LICENSE file

#ifndef OS_TIME_HPP
#define OS_TIME_HPP

#include <time.h> // timespec, clock_gettime
#include <iosfwd> // std::ostream&

#if defined (_MSC_VER)
#define WIN32_LEAN_AND_MEAN
#include <windows.h> // GetSystemTimeAsFileTime
#endif

namespace OS
{

#if defined (_MSC_VER)
  struct timespec
  {
    time_t tv_sec; // seconds
    long tv_nsec;  // nanoseconds
  } timespec_t;
#else
  typedef ::timespec timespec;
#endif

  //@@ what should be an initial value when constructed with
  //   default ctor? Options so far: {0, current time}
  //
  //@@ should I provide conversion to ::timespec const& and
  //   ::timespec const*
  //
  class Time
  {
  public:
    class Failed {};

    Time ()
    {
#if defined (_MSC_VER)
      FILETIME ft;
      GetSystemTimeAsFileTime (&ft);
      unsigned long long v ((unsigned long long (ft.dwHighDateTime) << 32) +
        ft.dwLowDateTime);

      ts_.tv_sec  = static_cast<time_t> (v / 10000000ULL);
      ts_.tv_nsec = static_cast<long> ((v % 10000000ULL) * 100);

#else
      if (clock_gettime (CLOCK_REALTIME /*CLOCK_MONOTONIC*/, &ts_) != 0)
      {
        throw Failed ();
      }
#endif
    }

    Time (unsigned long long nsec)
    {
      ts_.tv_sec  = static_cast<time_t> (nsec / 1000000000ULL);
      ts_.tv_nsec = static_cast<long> (nsec % 1000000000ULL);
    }

    Time (unsigned long sec, unsigned long nsec)
    {
      ts_.tv_sec  = sec;
      ts_.tv_nsec = nsec;
    }

  public:
    time_t
    sec () const
    {
      return ts_.tv_sec;
    }

    long
    nsec () const
    {
      return ts_.tv_nsec;
    }

  public:
    OS::timespec const&
    timespec () const
    {
      return ts_;
    }

  public:

    class Overrflow {};
    class Underflow {};


    // Member operators
    //
    Time
    operator+= (Time const& b)
    {
      //@@ overflow is not detected yet

      unsigned long long tmp = 0ULL + ts_.tv_nsec + b.ts_.tv_nsec;

      ts_.tv_sec += static_cast<time_t> (b.ts_.tv_sec + tmp / 1000000000ULL);
      ts_.tv_nsec = static_cast<long> (tmp % 1000000000ULL);

      return *this;
    }

    Time
    operator-= (Time const& b)
    {
      if (*this < b) throw Underflow ();

      ts_.tv_sec -= b.ts_.tv_sec;

      if (ts_.tv_nsec < b.ts_.tv_nsec)
      {
        --ts_.tv_sec;
        ts_.tv_nsec += 1000000000ULL - b.ts_.tv_nsec;
      }
      else
      {
        ts_.tv_nsec -= b.ts_.tv_nsec;
      }

      return *this;
    }

    // Friendly oprators
    //
    friend Time
    operator+ (Time const& a, Time const& b)
    {
      Time r (a);
      r += b;
      return r;
    }

    friend Time
    operator- (Time const& a, Time const& b)
    {
      Time r (a);
      r -= b;
      return r;
    }

    friend bool
    operator < (Time const& a, Time const& b)
    {
      return
        (a.ts_.tv_sec < b.ts_.tv_sec) ||
        (a.ts_.tv_sec == b.ts_.tv_sec && a.ts_.tv_nsec < b.ts_.tv_nsec);
    }

  private:
    OS::timespec ts_;
  };

  std::ostream&
  operator<< (std::ostream&, Time const&);
}

#include <ostream> // std::ostream
#include <iomanip> // std::setfill, std::setw

namespace OS
{
  std::ostream&
  operator<< (std::ostream& o, Time const& t)
  {
    return o << t.sec () << ':'
             << std::setfill ('0') << std::setw (9) << t.nsec ();
  }
}

#endif // OS_TIME_HPP
